@@ -22,6 +22,7 @@ limitations under the License.
 
 #if PY_VERSION_HEX >= 0x02060000
 #include "bytesobject.h"
+#include "structseq.h"
 #elif PY_VERSION_HEX < 0x02060000
 #define PyBytes_AsString PyString_AsString
 #define PyBytes_Check PyString_Check
@@ -417,10 +418,26 @@ typedef struct _CALLBACK_DATA
   PyObject* modules_data;
   PyObject* modules_callback;
   PyObject* warnings_callback;
+  PyObject* console_callback;
   int which;
 
 } CALLBACK_DATA;
 
+static PyStructSequence_Field RuleString_Fields[] = {
+  {"namespace", "Namespace of the rule"},
+  {"rule", "Identifier of the rule"},
+  {"string", "Identifier of the string"},
+  {NULL}
+};
+
+static PyStructSequence_Desc RuleString_Desc = {
+  "RuleString",
+  "Named tuple tying together rule identifier and string identifier",
+  RuleString_Fields,
+  (sizeof(RuleString_Fields) / sizeof(RuleString_Fields[0])) - 1
+};
+
+static PyTypeObject RuleString_Type = {0};
 
 // Forward declarations for handling module data.
 PyObject* convert_structure_to_python(
@@ -675,6 +692,58 @@ static int handle_module_imported(
 }
 
 
+static int handle_console_log(
+    void* message_data,
+    CALLBACK_DATA* data)
+{
+  PyGILState_STATE gil_state = PyGILState_Ensure();
+  int result = CALLBACK_CONTINUE;
+
+  if (data->console_callback == NULL)
+  {
+    // If the user does not specify a console callback we dump to stdout.
+    // If we want to support 3.2 and newer only we can use
+    // https://docs.python.org/3/c-api/sys.html?highlight=stdout#c.PySys_FormatStdout
+    // instead of this call with the limit.
+    PySys_WriteStdout("%.1000s\n", (char*) message_data);
+  }
+  else
+  {
+    PyObject* log_string = PY_STRING((char*) message_data);
+    Py_INCREF(data->console_callback);
+
+    PyObject* callback_result = PyObject_CallFunctionObjArgs(
+        data->console_callback,
+        log_string,
+        NULL);
+
+    if (callback_result != NULL)
+    {
+      #if PY_MAJOR_VERSION >= 3
+      if (PyLong_Check(callback_result))
+      #else
+      if (PyLong_Check(callback_result) || PyInt_Check(callback_result))
+      #endif
+      {
+        result = (int) PyLong_AsLong(callback_result);
+      }
+    }
+    else
+    {
+      result = CALLBACK_ERROR;
+    }
+
+    Py_DECREF(log_string);
+    Py_XDECREF(callback_result);
+    Py_DECREF(data->console_callback);
+  }
+
+  PyGILState_Release(gil_state);
+
+  return result;
+}
+
+
 static int handle_too_many_matches(
     YR_SCAN_CONTEXT* context,
     YR_STRING* string,
@@ -683,7 +752,11 @@ static int handle_too_many_matches(
   PyGILState_STATE gil_state = PyGILState_Ensure();
 
   PyObject* warning_type = NULL;
-  PyObject* identifier = NULL;
+  PyObject* string_identifier = NULL;
+  PyObject* rule_identifier = NULL;
+  PyObject* namespace_identifier = NULL;
+  PyObject* rule_string = NULL;
+  YR_RULE* rule = NULL;
 
   int result = CALLBACK_CONTINUE;
 
@@ -705,14 +778,49 @@ static int handle_too_many_matches(
   {
     Py_INCREF(data->warnings_callback);
 
-    identifier = PyBytes_FromString(string->identifier);
+    string_identifier = PY_STRING(string->identifier);
 
-    if (identifier == NULL)
+    if (string_identifier == NULL)
     {
       result = CALLBACK_ERROR;
       goto _exit;
     }
 
+    rule = &context->rules->rules_table[string->rule_idx];
+    rule_identifier = PY_STRING(rule->identifier);
+
+    if (rule_identifier == NULL)
+    {
+      result = CALLBACK_ERROR;
+      goto _exit;
+    }
+
+    namespace_identifier = PY_STRING(rule->ns->name);
+
+    if (namespace_identifier == NULL)
+    {
+      result = CALLBACK_ERROR;
+      goto _exit;
+    }
+
+    rule_string = PyStructSequence_New(&RuleString_Type);
+
+    if (rule_string == NULL)
+    {
+      result = CALLBACK_ERROR;
+      goto _exit;
+    }
+
+    PyStructSequence_SET_ITEM(rule_string, 0, namespace_identifier);
+    PyStructSequence_SET_ITEM(rule_string, 1, rule_identifier);
+    PyStructSequence_SET_ITEM(rule_string, 2, string_identifier);
+
+    // PyStructSequenece steals the reference so we NULL these
+    // so that Py_XDECREF() can be used in _exit label
+    namespace_identifier = NULL;
+    rule_identifier = NULL;
+    string_identifier = NULL;
+
     warning_type = PyLong_FromLong(CALLBACK_MSG_TOO_MANY_MATCHES);
 
     if (warning_type == NULL)
@@ -724,7 +832,7 @@ static int handle_too_many_matches(
     PyObject* callback_result = PyObject_CallFunctionObjArgs(
         data->warnings_callback,
         warning_type,
-        identifier,
+        rule_string,
         NULL);
 
     if (callback_result != NULL)
@@ -748,7 +856,10 @@ static int handle_too_many_matches(
 
 _exit:
 
-  Py_XDECREF(identifier);
+  Py_XDECREF(namespace_identifier);
+  Py_XDECREF(rule_identifier);
+  Py_XDECREF(string_identifier);
+  Py_XDECREF(rule_string);
   Py_XDECREF(warning_type);
   Py_XDECREF(data->warnings_callback);
 
@@ -812,6 +923,10 @@ int yara_callback(
     if (callback == NULL ||
         (which & CALLBACK_NON_MATCHES) != CALLBACK_NON_MATCHES)
       return CALLBACK_CONTINUE;
+    break;
+
+  case CALLBACK_MSG_CONSOLE_LOG:
+    return handle_console_log(message_data, user_data);
   }
 
   // At this point we have handled all the other cases of when this callback
@@ -1174,7 +1289,7 @@ int process_compile_externals(
 
 int process_match_externals(
     PyObject* externals,
-    YR_RULES* rules)
+    YR_SCANNER* scanner)
 {
   PyObject* key;
   PyObject* value;
@@ -1189,8 +1304,8 @@ int process_match_externals(
 
     if (PyBool_Check(value))
     {
-      result = yr_rules_define_boolean_variable(
-          rules,
+      result = yr_scanner_define_boolean_variable(
+          scanner,
           identifier,
           PyObject_IsTrue(value));
     }
@@ -1200,15 +1315,15 @@ int process_match_externals(
     else if (PyLong_Check(value) || PyInt_Check(value))
 #endif
     {
-      result = yr_rules_define_integer_variable(
-          rules,
+      result = yr_scanner_define_integer_variable(
+          scanner,
           identifier,
           PyLong_AsLongLong(value));
     }
     else if (PyFloat_Check(value))
     {
-      result = yr_rules_define_float_variable(
-          rules,
+      result = yr_scanner_define_float_variable(
+          scanner,
           identifier,
           PyFloat_AsDouble(value));
     }
@@ -1219,8 +1334,8 @@ int process_match_externals(
       if (str == NULL)
         return ERROR_INVALID_ARGUMENT;
 
-      result = yr_rules_define_string_variable(
-          rules, identifier, str);
+      result = yr_scanner_define_string_variable(
+          scanner, identifier, str);
     }
     else
     {
@@ -1231,7 +1346,7 @@ int process_match_externals(
       return ERROR_INVALID_ARGUMENT;
     }
 
-    // yr_rules_define_xxx_variable returns ERROR_INVALID_ARGUMENT if the
+    // yr_scanner_define_xxx_variable returns ERROR_INVALID_ARGUMENT if the
     // variable wasn't previously defined in the compilation phase. Ignore
     // those errors because we don't want the "scan" method being aborted
     // because of the "externals" dictionary having more keys than those used
@@ -1493,7 +1608,8 @@ static PyObject* Rules_match(
   static char* kwlist[] = {
       "filepath", "pid", "data", "externals",
       "callback", "fast", "timeout", "modules_data",
-      "modules_callback", "which_callbacks", "warnings_callback", NULL
+      "modules_callback", "which_callbacks", "warnings_callback",
+      "console_callback", NULL
       };
 
   char* filepath = NULL;
@@ -1502,13 +1618,13 @@ static PyObject* Rules_match(
   int pid = -1;
   int timeout = 0;
   int error = ERROR_SUCCESS;
-  int fast_mode = 0;
 
   PyObject* externals = NULL;
   PyObject* fast = NULL;
 
   Rules* object = (Rules*) self;
 
+  YR_SCANNER* scanner;
   CALLBACK_DATA callback_data;
 
   callback_data.matches = NULL;
@@ -1516,12 +1632,13 @@ static PyObject* Rules_match(
   callback_data.modules_data = NULL;
   callback_data.modules_callback = NULL;
   callback_data.warnings_callback = NULL;
+  callback_data.console_callback = NULL;
   callback_data.which = CALLBACK_ALL;
 
   if (PyArg_ParseTupleAndKeywords(
         args,
         keywords,
-        "|sis*OOOiOOiO",
+        "|sis*OOOiOOiOO",
         kwlist,
         &filepath,
         &pid,
@@ -1533,7 +1650,8 @@ static PyObject* Rules_match(
         &callback_data.modules_data,
         &callback_data.modules_callback,
         &callback_data.which,
-        &callback_data.warnings_callback))
+        &callback_data.warnings_callback,
+        &callback_data.console_callback))
   {
     if (filepath == NULL && data.buf == NULL && pid == -1)
     {
@@ -1575,6 +1693,17 @@ static PyObject* Rules_match(
       }
     }
 
+    if (callback_data.console_callback != NULL)
+    {
+      if (!PyCallable_Check(callback_data.console_callback))
+      {
+        PyBuffer_Release(&data);
+        return PyErr_Format(
+            PyExc_TypeError,
+            "'console_callback' must be callable");
+      }
+    }
+
     if (callback_data.modules_data != NULL)
     {
       if (!PyDict_Check(callback_data.modules_data))
@@ -1586,46 +1715,49 @@ static PyObject* Rules_match(
       }
     }
 
+    if (yr_scanner_create(object->rules, &scanner) != 0)
+    {
+      return PyErr_Format(
+          PyExc_Exception,
+          "could not create scanner");
+    }
+
     if (externals != NULL && externals != Py_None)
     {
       if (PyDict_Check(externals))
       {
-        if (process_match_externals(externals, object->rules) != ERROR_SUCCESS)
+        if (process_match_externals(externals, scanner) != ERROR_SUCCESS)
         {
-          // Restore original externals provided during compiling.
-          process_match_externals(object->externals, object->rules);
-
           PyBuffer_Release(&data);
+          yr_scanner_destroy(scanner);
           return NULL;
         }
       }
       else
       {
         PyBuffer_Release(&data);
+        yr_scanner_destroy(scanner);
         return PyErr_Format(
             PyExc_TypeError,
             "'externals' must be a dictionary");
       }
     }
 
-    if (fast != NULL)
+    if (fast != NULL && PyObject_IsTrue(fast) == 1)
     {
-      fast_mode = (PyObject_IsTrue(fast) == 1);
+      yr_scanner_set_flags(scanner, SCAN_FLAGS_FAST_MODE);
     }
 
+    yr_scanner_set_timeout(scanner, timeout);
+    yr_scanner_set_callback(scanner, yara_callback, &callback_data);
+
     if (filepath != NULL)
     {
       callback_data.matches = PyList_New(0);
 
       Py_BEGIN_ALLOW_THREADS
 
-      error = yr_rules_scan_file(
-          object->rules,
-          filepath,
-          fast_mode ? SCAN_FLAGS_FAST_MODE : 0,
-          yara_callback,
-          &callback_data,
-          timeout);
+      error = yr_scanner_scan_file(scanner, filepath);
 
       Py_END_ALLOW_THREADS
     }
@@ -1635,14 +1767,10 @@ static PyObject* Rules_match(
 
       Py_BEGIN_ALLOW_THREADS
 
-      error = yr_rules_scan_mem(
-          object->rules,
+      error = yr_scanner_scan_mem(
+          scanner,
           (unsigned char*) data.buf,
-          (size_t) data.len,
-          fast_mode ? SCAN_FLAGS_FAST_MODE : 0,
-          yara_callback,
-          &callback_data,
-          timeout);
+          (size_t) data.len);
 
       Py_END_ALLOW_THREADS
     }
@@ -1652,29 +1780,13 @@ static PyObject* Rules_match(
 
       Py_BEGIN_ALLOW_THREADS
 
-      error = yr_rules_scan_proc(
-          object->rules,
-          pid,
-          fast_mode ? SCAN_FLAGS_FAST_MODE : 0,
-          yara_callback,
-          &callback_data,
-          timeout);
+      error = yr_scanner_scan_proc(scanner, pid);
 
       Py_END_ALLOW_THREADS
     }
 
     PyBuffer_Release(&data);
-
-    // Restore original externals provided during compiling.
-    if (object->externals != NULL)
-    {
-      if (process_match_externals(
-            object->externals, object->rules) != ERROR_SUCCESS)
-      {
-        Py_DECREF(callback_data.matches);
-        return NULL;
-      }
-    }
+    yr_scanner_destroy(scanner);
 
     if (error != ERROR_SUCCESS)
     {
@@ -1833,6 +1945,8 @@ void raise_exception_on_error(
     const char* message,
     void* user_data)
 {
+  PyGILState_STATE gil_state = PyGILState_Ensure();
+
   if (error_level == YARA_ERROR_LEVEL_ERROR)
   {
     if (file_name != NULL)
@@ -1867,6 +1981,8 @@ void raise_exception_on_error(
     PyList_Append(warnings, warning_msg);
     Py_DECREF(warning_msg);
   }
+
+  PyGILState_Release(gil_state);
 }
 
 
@@ -2083,6 +2199,27 @@ static PyObject* yara_compile(
         &error_on_warning,
         &include_callback))
   {
+    char num_args = 0;
+
+    if (filepath != NULL)
+      num_args++;
+
+    if (source != NULL)
+      num_args++;
+
+    if (file != NULL)
+      num_args++;
+
+    if (filepaths_dict != NULL)
+      num_args++;
+
+    if (sources_dict != NULL)
+      num_args++;
+
+    if (num_args > 1)
+      return PyErr_Format(
+          PyExc_TypeError,
+          "compile is receiving too many arguments");
 
     error = yr_compiler_create(&compiler);
 
@@ -2170,8 +2307,10 @@ static PyObject* yara_compile(
 
       if (fh != NULL)
       {
+        Py_BEGIN_ALLOW_THREADS
         error = yr_compiler_add_file(compiler, fh, NULL, filepath);
         fclose(fh);
+        Py_END_ALLOW_THREADS
       }
       else
       {
@@ -2180,7 +2319,9 @@ static PyObject* yara_compile(
     }
     else if (source != NULL)
     {
+      Py_BEGIN_ALLOW_THREADS
       error = yr_compiler_add_string(compiler, source, NULL);
+      Py_END_ALLOW_THREADS
     }
     else if (file != NULL)
     {
@@ -2188,9 +2329,11 @@ static PyObject* yara_compile(
 
       if (fd != -1)
       {
+        Py_BEGIN_ALLOW_THREADS
         fh = fdopen(fd, "r");
         error = yr_compiler_add_file(compiler, fh, NULL, NULL);
         fclose(fh);
+        Py_END_ALLOW_THREADS
       }
       else
       {
@@ -2210,7 +2353,9 @@ static PyObject* yara_compile(
 
           if (source != NULL && ns != NULL)
           {
+            Py_BEGIN_ALLOW_THREADS
             error = yr_compiler_add_string(compiler, source, ns);
+            Py_END_ALLOW_THREADS
 
             if (error > 0)
               break;
@@ -2247,8 +2392,10 @@ static PyObject* yara_compile(
 
             if (fh != NULL)
             {
+              Py_BEGIN_ALLOW_THREADS
               error = yr_compiler_add_file(compiler, fh, ns, filepath);
               fclose(fh);
+              Py_END_ALLOW_THREADS
 
               if (error > 0)
                 break;
@@ -2563,6 +2710,8 @@ MOD_INIT(yara)
   if (PyType_Ready(&Match_Type) < 0)
     return MOD_ERROR_VAL;
 
+  PyStructSequence_InitType(&RuleString_Type, &RuleString_Desc);
+
   PyModule_AddObject(m, "Rule", (PyObject*) &Rule_Type);
   PyModule_AddObject(m, "Rules", (PyObject*) &Rules_Type);
   PyModule_AddObject(m, "Match",  (PyObject*) &Match_Type);
